#pragma once
#include <vector>
#include <string>
#include <atomic>
#include "Utils.h"
#include "ffgl/FFGL.h"

// 2019-07-17 ffgl@subpixels.com
// - Replaced "Ptr" types (shared_ptr) with regular pointers
// - Removed create_x() functions
// - Expand constructor list for char* and sting& parameters

/*
typedef struct ParamInfoStruct
{
	unsigned int ID;
	char Name[ 16 ];
	ParameterType type;

	// extra parameters
	struct Element
	{
		std::string name;
		float value = 0.0f;
	};
	std::vector< Element > elements;
	ParameterUsage usage;

	RangeStruct range;
	float DefaultValue;
	char* StrDefaultValue;
	ParamInfoStruct* pNext;

	ParamInfoStruct() :
		range{ 0, 1 }
	{
	}
} ParamInfo;
*/

class Param
{
protected:
	unsigned int index;

	std::string name;
	ffgl::ParameterType type;
	float value = 0;

	std::string uniformName;
	GLint uniformLocation;

public:
	Param() :
		Param( std::string() )
	{
	}

	Param( const std::string& name ) :
		Param( name, 0.0f )
	{
	}
	Param( const char* name ) :
		Param( name, 0.0f )
	{
	}

	Param( const std::string& name, float value ) :
		Param( name, ffgl::ParameterType::FF_TYPE_STANDARD, value )
	{
	}
	Param( const char* name, float value ) :
		Param( name, ffgl::ParameterType::FF_TYPE_STANDARD, value )
	{
	}

	Param( const std::string& name, ffgl::ParameterType type, float value ) :
		Param( name, type, value, name )
	{
	}
	Param( const char* name, ffgl::ParameterType type, float value ) :
		Param( name, type, value, name )
	{
	}

	Param( const std::string& name, ffgl::ParameterType type, float value, const std::string& uniformName ) :
		name( name ),
		type( type ),
		value( value ),
		uniformName( uniformName ),
		uniformLocation( -1 )
	{
	}
	Param( const char* name, ffgl::ParameterType type, float value, const char* uniformName ) :
		name( name ),
		type( type ),
		value( value ),
		uniformName( uniformName ),
		uniformLocation( -1 )
	{
	}

	virtual ~Param()
	{
	}

	virtual void setIndex( unsigned int ID )
	{
		index = ID;
	}
	virtual unsigned int getIndex() const
	{
		return index;
	}
	const std::string& getName() const
	{
		return name;
	}
	ffgl::ParameterType getType() const
	{
		return type;
	}
	virtual void setValue( float v )
	{
		value = v;
	}
	virtual float getValue() const
	{
		return value;
	}
	const std::string& getUniformName() const
	{
		return uniformName;
	}
	void setUniformLocation(GLint location)
	{
		uniformLocation = location;
	}
	const GLint getUniformLocation() const
	{
		return uniformLocation;
	}
};

class ParamRange : public Param
{
public:
	struct Range
	{
		float min = 0.0f;
		float max = 1.0f;

		Range( float min, float max ) :
			min( min ),
			max( max ){};
	};

	ParamRange( std::string name, float value, Range range ) :
		Param( name, ffgl::ParameterType::FF_TYPE_STANDARD, value ),
		range( range )
	{
		setValue( utils::map( value, range.min, range.max, 0.0f, 1.0f ) );
	}

	float getRealValue()
	{
		return utils::map( value, 0.0f, 1.0f, range.min, range.max );
	}

	Range getRange()
	{
		return range;
	}

protected:
	Range range;
};

class ParamInteger : public ParamRange
{
public:
	ParamInteger( std::string name, int value, Range range ) :
		ParamRange( name, (float)value, range)
	{
		type = ffgl::ParameterType::FF_TYPE_INTEGER;
	}
};

class ParamOption : public Param
{
public:
	struct Option
	{
		std::string name;
		float value;
	};
	typedef std::vector< Option > Options;
	Options options;

	ParamOption( std::string name, Options options ) :
		ParamOption( name, options, 0 )
	{
	}
	ParamOption( std::string name, Options options, int defaultOption ) :
		Param( name, ffgl::ParameterType::FF_TYPE_OPTION, (float)defaultOption ),
		options( options )
	{
		setValue( (float)defaultOption );
	}

	void setValue( float _value ) override
	{
		value = options.size() <= _value ? 0 : _value;
		currentOption = options[ (size_t)value ];
	}

	float getRealValue()
	{
		return currentOption.value;
	}

	bool isCurrentOption( std::string option )
	{
		return option.compare( currentOption.name ) == 0;
	}

private:
	Option currentOption;
};

class ParamEvent : public Param
{
public:
	ParamEvent( std::string name ) :
		Param( name, ffgl::ParameterType::FF_TYPE_EVENT, 0 )
	{
	}
};

class ParamTrigger : public ParamEvent
{
public:
	ParamTrigger( std::string name ) :
		ParamEvent( name )
	{
	}

	void consume()
	{
		value = false;
	}
};

class ParamBool : public Param
{
public:
	ParamBool( std::string name ) :
		ParamBool( name, false, name )
	{
	}
	ParamBool( std::string name, bool defaultValue ) :
		ParamBool( name, defaultValue, name )
	{
	}
	ParamBool( std::string name, std::string uniformName ) :
		ParamBool( name, false, uniformName )
	{
	}
	ParamBool( std::string name, bool defaultValue, std::string uniformName ) :
		Param( name, ffgl::ParameterType::FF_TYPE_BOOLEAN, defaultValue, uniformName )
	{
	}
};

class ParamText : public Param
{
public:
	ParamText( std::string& name ) :
		ParamText( name, std::string() )
	{
	}
	ParamText( const char* name ) :
		ParamText( name, "" )
	{
	}

	ParamText( std::string& name, std::string& text ) :
		Param( name, ffgl::ParameterType::FF_TYPE_TEXT, 0 ),
		text( text )
	{
	}
	ParamText( const char* name, const char* text ) :
		Param( name, ffgl::ParameterType::FF_TYPE_TEXT, 0 ),
		text( text )
	{
	}

	// 2019-07-17 ffgl@subpixels.com - Create text setter
	void setText( const std::string& text )
	{
		this->text = text;
	}
	// 2019-07-17 ffgl@subpixels.com - Create text getter
	const std::string& getText() const
	{
		return text;
	}
	// 2019-07-17 ffgl@subpixels.com - Make text protected
protected:
	std::string text;
};

class ParamFFT : public Param
{
	// 2019-07-22 ffgl@subpixels.com - Reduce default buffer size (number of FFT bins) from 2048 to 16
	static const int DEFAULT_BUFFER_SIZE = 16;

public:
	ParamFFT( std::string name, size_t size = DEFAULT_BUFFER_SIZE ) :
		Param( name ),
		fftData( size )
	{
		type = ffgl::ParameterType::FF_TYPE_BUFFER;
	}

	std::vector< float > fftData;
};
